from autofit import conf
from autolens.data import ccd
from autolens.data.array import mask as msk
from autolens.data.plotters import ccd_plotters
from autolens.lens import lens_data as ld
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.galaxy import galaxy as g
from autolens.lens import ray_tracing
from autolens.data.plotters import ccd_plotters
from autolens.lens.plotters import plane_plotters

import os

path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
af.conf.instance = af.conf.Config(
    config_path=path + "config", output_path=path + "output"
)

lens_name = "slacs0737+3216"  # Works

ccd_data = ccd.load_ccd_data_from_fits(
    image_path=path + "/data/slacs/" + lens_name + "/F814W_image.fits",
    psf_path=path + "/data/slacs/" + lens_name + "/F814W_psf.fits",
    noise_map_path=path + "/data/slacs/" + lens_name + "/F814W_noise_map.fits",
    pixel_scale=0.03,
    resized_ccd_shape=(301, 301),
    resized_psf_shape=(21, 21),
)

mask = msk.load_mask_from_fits(
    mask_path=path + "/data/slacs/" + lens_name + "/mask.fits", pixel_scale=0.03
)

# ccd_plotters.plot_image(ccd_data=ccd_data, mask=mask)

lens_data = ld.LensData(ccd_data=ccd_data, mask=mask)

sie = mp.EllipticalIsothermal(
    centre=(0.007332219831800636, -0.017091734979904037),
    axis_ratio=0.8394667884862433,
    einstein_radius=0.9733056210502408,
    phi=97.61311065689567,
)
shear = mp.ExternalShear(magnitude=0.06821354969124815, phi=0.3748585976691017)

lens_galaxy = g.Galaxy(mass=sie, shear=shear)

source_galaxy = g.Galaxy(
    light=lp.EllipticalSersic(
        centre=(0.028673479294845223, -0.13237833417289763),
        axis_ratio=0.6681602590613542,
        phi=40.19143668987858,
        intensity=0.006995496893309517,
        effective_radius=0.389050969603033,
        sersic_index=2.965580682063875,
    )
)

tracer = ray_tracing.TracerImageSourcePlanes(
    lens_galaxies=[lens_galaxy],
    source_galaxies=[g.Galaxy(light=lp.SphericalSersic(intensity=0.0))],
    image_plane_grid_stack=lens_data.grid_stack,
)

path = "{}/".format(os.path.dirname(os.path.realpath(__file__)))

plane_plotters.plot_plane_image(
    plane=tracer.source_plane,
    positions=None,
    plot_grid=True,
    units="arcsec",
    figsize=(7, 7),
    aspect="auto",
    output_path=path,
    output_filename="grid_only",
    output_format="png",
)

tracer = ray_tracing.TracerImageSourcePlanes(
    lens_galaxies=[lens_galaxy],
    source_galaxies=[source_galaxy],
    image_plane_grid_stack=lens_data.grid_stack,
)

plane_plotters.plot_plane_image(
    plane=tracer.source_plane,
    positions=None,
    plot_grid=True,
    units="arcsec",
    figsize=(7, 7),
    aspect="auto",
    output_path=path,
    output_filename="source_and_grid",
    output_format="png",
)

plane_plotters.plot_plane_image(
    plane=tracer.source_plane,
    positions=None,
    plot_grid=False,
    units="arcsec",
    figsize=(7, 7),
    aspect="auto",
    output_path=path,
    output_filename="source_no_grid",
    output_format="png",
)



import autofit as af
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline
from autolens.pipeline import tagging as tag


def make_pipeline(phase_folders=None):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pl__complex_source"

    pipeline_name = tag.pipeline_name_from_name_and_settings(
        pipeline_name=pipeline_name
    )

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/phase_name/'
    phase_folders.append(pipeline_name)

    # To begin, we need to initialize the lens's mass model. We should be able to do this by using a simple source
    # model. It won't fit the complicated structure of the source, but it'll give us a reasonable estimate of the
    # einstein radius and the other lens-mass parameters.

    # This should run fine without any prior-passes. In general, a thick, giant ring of source light is something we
    # can be confident MultiNest will fit without much issue, especially when the lens galaxy's light isn't included
    # such that the parameter space is just 12 parameters.

    phase1 = phase_imaging.LensSourcePlanePhase(
        phase_name="phase_1_simple_source",
        phase_folders=phase_folders,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(redshift=0.5, mass=mp.EllipticalIsothermal)
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(redshift=1.0, light_0=lp.EllipticalSersic)
        ),
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.5

    # Now lets add another source component, using the previous model as the initialization on the lens / source
    # parameters. We'll vary the parameters of the lens mass model and first source galaxy component during the fit.

    class X2SourcePhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            self.lens_galaxies.lens = results.from_phase(
                "phase_1_simple_source"
            ).variable.lens

            self.source_galaxies.source.light_0 = results.from_phase(
                "phase_1_simple_source"
            ).variable.source.light_0

    # You'll notice I've stop writing 'phase_1_results = results.from_phase('phase_1_simple_source')' - we know how
    # the previous results are structured now so lets not clutter our code!

    phase2 = X2SourcePhase(
        phase_name="phase_2_x2_source",
        phase_folders=phase_folders,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(redshift=0.5, mass=mp.EllipticalIsothermal)
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(
                redshift=1.0,
                light_0=lp.EllipticalExponential,
                light_1=lp.EllipticalSersic,
            )
        ),
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 40
    phase2.optimizer.sampling_efficiency = 0.5

    # Now lets do the same again, but with 3 source galaxy components.

    class X3SourcePhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            self.lens_galaxies.lens = results.from_phase(
                "phase_2_x2_source"
            ).variable.lens_galaxies.lens

            self.source_galaxies.source.light_0 = results.from_phase(
                "phase_2_x2_source"
            ).variable.source_galaxies.source.light_0

            self.source_galaxies.source.light_1 = results.from_phase(
                "phase_2_x2_source"
            ).variable.source_galaxies.source.light_1

    phase3 = X3SourcePhase(
        phase_name="phase_3_x3_source",
        phase_folders=phase_folders,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(redshift=0.5, mass=mp.EllipticalIsothermal)
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(
                redshift=1.0,
                light_0=lp.EllipticalExponential,
                light_1=lp.EllipticalSersic,
                light_2=lp.EllipticalSersic,
            )
        ),
        optimizer_class=af.MultiNest,
    )

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 50
    phase3.optimizer.sampling_efficiency = 0.5

    # And one more for luck!

    class X4SourcePhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            self.lens_galaxies.lens = results.from_phase(
                "phase_3_x3_source"
            ).variable.lens_galaxies.lens

            self.source_galaxies.source.light_0 = results.from_phase(
                "phase_3_x3_source"
            ).variable.source_galaxies.source.light_0

            self.source_galaxies.source.light_1 = results.from_phase(
                "phase_3_x3_source"
            ).variable.source_galaxies.source.light_1

            self.source_galaxies.source.light_2 = results.from_phase(
                "phase_3_x3_source"
            ).variable.source_galaxies.source.light_2

    phase4 = X4SourcePhase(
        phase_name="phase_4_x4_source",
        phase_folders=phase_folders,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(redshift=0.5, mass=mp.EllipticalIsothermal)
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(
                redshift=1.0,
                light_0=lp.EllipticalExponential,
                light_1=lp.EllipticalSersic,
                light_2=lp.EllipticalSersic,
                light_3=lp.EllipticalSersic,
            )
        ),
        optimizer_class=af.MultiNest,
    )

    phase4.optimizer.const_efficiency_mode = True
    phase4.optimizer.n_live_points = 50
    phase4.optimizer.sampling_efficiency = 0.5

    return pipeline.PipelineImaging(pipeline_name, phase1, phase2, phase3, phase4)



import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging, phase_extensions
from autolens.pipeline import pipeline
from autolens.pipeline import tagging as tag
from autolens.model.profiles import mass_profiles as mp
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg

import os

# In this pipeline, we'll perform a subhalo analysis which determines the sensitivity map of a strong lens and
# then attempts to detection subhalos by putting subhalos at fixed intevals on a 2D (y,x) grid. The source uses an
# inversion. The pipeline is as follows:

# Phase 1:

# Description: Perform the subhalo detection analysis.
# Lens Mass: EllipitcalPowerLaw + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Subhalo: SphericalTruncatedNFWChallenge
# Previous Pipelines: initialize/lens_sie_shear_source_inversion_from_pipeline.py
# Prior Passing: Lens Mass constant from previous pipeline, source light variable from previous pipeline.
# Notes: Priors on subhalo are tuned to give realistic masses (10^6 - 10^10) and concentrations (6-24)

# Phase 2:

# Description: Refine the best-fit detected subhalo from the previous phase, by varying also the lens mass model.
# Lens Mass: EllipitcalPowerLaw + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Subhalo: SphericalTruncatedNFWChallenge
# Previous Pipelines: initialize/lens_sie_shear_source_inversion_from_pipeline.py
# Prior Passing: Lens Mass variable from previous pipeline, source light and subhalo mass variable from phase 2.
# Notes: None


def make_pipeline(
    pl_pixelization=pix.VoronoiMagnification,
    pl_regularization=reg.AdaptiveBrightness,
    phase_folders=None,
    tag_phases=True,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    interp_pixel_scale=0.05,
    inversion_pixel_limit=None,
    cluster_pixel_scale=0.1,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_subhalo__lens_pl_shear_subhalo_source_inversion"

    pipeline_name = tag.pipeline_name_from_name_and_settings(
        pipeline_name=pipeline_name,
        pixelization=pl_pixelization,
        regularization=pl_regularization,
    )

    phase_folders.append(pipeline_name)

    ### Phase 1 ###

    # In phase 1, we attempt to detect subhalos, by performing a NxN grid search of MultiNest searches, where:

    # 1) The lens model and source-pixelization parameters are held fixed to the best-fit values from phase 2.
    # 2) Each grid search varies the subhalo (y,x) coordinates and mass as free parameters.
    # 3) The priors on these (y,x) coordinates are UniformPriors, with limits corresponding to the grid-cells.

    class GridPhase(af.phase.as_grid_search(phase_imaging.LensSourcePlanePhase)):
        @property
        def grid_priors(self):
            return [
                self.variable.lens_galaxies.subhalo.mass.centre_0,
                self.variable.lens_galaxies.subhalo.mass.centre_1,
            ]

        def pass_priors(self, results):

            ### Lens Mass, PL -> PL, Shear -> Shear ###

            self.lens_galaxies.lens = results.from_phase(
                "phase_1_lens_pl_shear_source_inversion"
            ).constant.lens_galaxies.lens

            ### Lens Subhalo, Adjust priors to physical masses (10^6 - 10^10) and concentrations (6-24)

            self.lens_galaxies.subhalo.mass.kappa_s = af.UniformPrior(
                lower_limit=0.0005, upper_limit=0.2
            )
            self.lens_galaxies.subhalo.mass.scale_radius = af.UniformPrior(
                lower_limit=0.001, upper_limit=1.0
            )
            self.lens_galaxies.subhalo.mass.centre_0 = af.UniformPrior(
                lower_limit=-2.0, upper_limit=2.0
            )
            self.lens_galaxies.subhalo.mass.centre_1 = af.UniformPrior(
                lower_limit=-2.0, upper_limit=2.0
            )

            ### Source Inversion, Inv -> Inv ###

            self.source_galaxies.source = results.from_phase(
                "phase_1_lens_pl_shear_source_inversion"
            ).inversion.constant.source_galaxies.source

    phase1 = GridPhase(
        phase_name="phase_1_subhalo_search",
        phase_folders=phase_folders,
        tag_phases=tag_phases,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                mass=mp.EllipticalPowerLaw,
                shear=mp.ExternalShear,
            ),
            subhalo=gm.GalaxyModel(
                redshift=redshift_lens, mass=mp.SphericalTruncatedNFWChallenge
            ),
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(
                redshift=redshift_source,
                pixelization=pl_pixelization,
                regularization=pl_regularization,
            )
        ),
        sub_grid_size=sub_grid_size,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        nterp_pixel_scale=interp_pixel_scale,
        inversion_pixel_limit=inversion_pixel_limit,
        cluster_pixel_scale=cluster_pixel_scale,
        optimizer_class=af.MultiNest,
        number_of_steps=4,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 20
    phase1.optimizer.sampling_efficiency = 0.3

    class SubhaloPhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            ### Lens Mass, PL -> PL, Shear -> Shear ###

            self.lens_galaxies.lens = results.from_phase(
                "phase_1_lens_pl_shear_source_inversion"
            ).variable.lens_galaxies.lens

            ### Subhalo, TruncatedNFW -> TruncatedNFW ###

            self.lens_galaxies.subhalo.mass.centre = (
                results.from_phase("phase_1_subhalo_search")
                .best_result.variable_absolute(a=0.3)
                .lens_galaxies.subhalo.mass.centre
            )

            self.lens_galaxies.subhalo.mass.kappa_s = (
                results.from_phase("phase_1_subhalo_search")
                .best_result.variable_relative(r=0.5)
                .lens_galaxies.subhalo.mass.kappa_s
            )

            self.lens_galaxies.subhalo.mass.scale_radius = (
                results.from_phase("phase_1_subhalo_search")
                .best_result.variable_relative(r=0.5)
                .lens_galaxies.subhalo.mass.scale_radius
            )

            ### Source Inversion, Inv -> Inv ###

            self.source_galaxies.source = results.from_phase(
                "phase_1_lens_pl_shear_source_inversion"
            ).inversion.constant.source_galaxies.source

    phase2 = SubhaloPhase(
        phase_name="phase_2_subhalo_refine",
        phase_folders=phase_folders,
        tag_phases=tag_phases,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                mass=mp.EllipticalPowerLaw,
                shear=mp.ExternalShear,
            ),
            subhalo=gm.GalaxyModel(
                redshift=redshift_lens, mass=mp.SphericalTruncatedNFWChallenge
            ),
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(
                redshift=redshift_source,
                pixelization=pl_pixelization,
                regularization=pl_regularization,
            )
        ),
        sub_grid_size=sub_grid_size,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        interp_pixel_scale=interp_pixel_scale,
        inversion_pixel_limit=inversion_pixel_limit,
        cluster_pixel_scale=cluster_pixel_scale,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 80
    phase2.optimizer.sampling_efficiency = 0.3

    return pipeline.PipelineImaging(pipeline_name, phase1, phase2)



import autofit as af


class PipelineImaging(af.Pipeline):
    def run(self, data, mask=None, positions=None, data_name=None):
        def runner(phase, results):
            return phase.run(data=data, results=results, mask=mask, positions=positions)

        return self.run_function(runner, data_name)


class PipelinePositions(af.Pipeline):
    def run(self, positions, pixel_scale):
        def runner(phase, results):
            return phase.run(
                positions=positions, pixel_scale=pixel_scale, results=results
            )

        return self.run_function(runner)



import os

import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.model.profiles import light_profiles as lp
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline as pl
from test.integration import integration_util
from test.simulation import simulation_util

test_type = "grid_search"
test_name = "multinest_grid_fixed_disk"

test_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
output_path = test_path + "output/"
config_path = test_path + "config"
af.conf.instance = af.conf.Config(config_path=config_path, output_path=output_path)


def pipeline():

    integration_util.reset_paths(test_name=test_name, output_path=output_path)
    ccd_data = simulation_util.load_test_ccd_data(
        data_type="lens_only_dev_vaucouleurs", data_resolution="Euclid"
    )
    pipeline = make_pipeline(test_name=test_name)
    pipeline.run(data=ccd_data)


def make_pipeline(test_name):
    class QuickPhase(phase_imaging.LensPlanePhase):
        def pass_priors(self, results):

            self.lens_galaxies.lens.bulge.centre_0 = af.UniformPrior(
                lower_limit=-0.01, upper_limit=0.01
            )
            self.lens_galaxies.lens.bulge.centre_1 = af.UniformPrior(
                lower_limit=-0.01, upper_limit=0.01
            )
            self.lens_galaxies.lens.bulge.axis_ratio = af.UniformPrior(
                lower_limit=0.79, upper_limit=0.81
            )
            self.lens_galaxies.lens.bulge.phi = af.UniformPrior(
                lower_limit=-1.0, upper_limit=1.0
            )
            self.lens_galaxies.lens.bulge.intensity = af.UniformPrior(
                lower_limit=0.99, upper_limit=1.01
            )
            self.lens_galaxies.lens.bulge.effective_radius = af.UniformPrior(
                lower_limit=1.25, upper_limit=1.35
            )
            self.lens_galaxies.lens.bulge.sersic_index = af.UniformPrior(
                lower_limit=3.95, upper_limit=4.05
            )

            self.lens_galaxies.lens.disk.centre_0 = af.UniformPrior(
                lower_limit=-0.01, upper_limit=0.01
            )
            self.lens_galaxies.lens.disk.centre_1 = af.UniformPrior(
                lower_limit=-0.01, upper_limit=0.01
            )
            self.lens_galaxies.lens.disk.axis_ratio = af.UniformPrior(
                lower_limit=0.69, upper_limit=0.71
            )
            self.lens_galaxies.lens.disk.phi = af.UniformPrior(
                lower_limit=-1.0, upper_limit=1.0
            )
            self.lens_galaxies.lens.disk.intensity = af.UniformPrior(
                lower_limit=1.99, upper_limit=2.01
            )
            self.lens_galaxies.lens.disk.effective_radius = af.UniformPrior(
                lower_limit=1.95, upper_limit=2.05
            )

    phase1 = QuickPhase(
        phase_name="phase_1",
        phase_folders=[test_type, test_name],
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=0.5, bulge=lp.EllipticalSersic, disk=lp.EllipticalSersic
            )
        ),
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.8

    class GridPhase(af.phase.as_grid_search(phase_imaging.LensPlanePhase)):
        @property
        def grid_priors(self):
            return [self.variable.lens_galaxies.lens.bulge.sersic_index]

        def pass_priors(self, results):

            self.lens_galaxies.lens.disk = results.from_phase(
                "phase_1"
            ).constant.lens_galaxies.lens.disk

            self.lens_galaxies.lens.bulge.centre_0 = af.UniformPrior(
                lower_limit=-0.01, upper_limit=0.01
            )
            self.lens_galaxies.lens.bulge.centre_1 = af.UniformPrior(
                lower_limit=-0.01, upper_limit=0.01
            )
            self.lens_galaxies.lens.bulge.axis_ratio = af.UniformPrior(
                lower_limit=0.79, upper_limit=0.81
            )
            self.lens_galaxies.lens.bulge.phi = af.UniformPrior(
                lower_limit=-1.0, upper_limit=1.0
            )
            self.lens_galaxies.lens.bulge.intensity = af.UniformPrior(
                lower_limit=0.99, upper_limit=1.01
            )
            self.lens_galaxies.lens.bulge.effective_radius = af.UniformPrior(
                lower_limit=1.25, upper_limit=1.35
            )

    phase2 = GridPhase(
        phase_name="phase_2",
        phase_folders=[test_type, test_name],
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=0.5, bulge=lp.EllipticalSersic, disk=lp.EllipticalSersic
            )
        ),
        number_of_steps=2,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True

    return pl.PipelineImaging(test_name, phase1, phase2)


if __name__ == "__main__":
    pipeline()



import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging, phase_extensions
from autolens.pipeline import pipeline
from autolens.pipeline import tagging as tag
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.profiles import light_and_mass_profiles as lmp
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg

import os


# In this pipeline, we'll perform an analysis which fits an image with the lens light included, and a source galaxy
# using an inversion, using a decomposed light and dark matter profile. The pipeline follows on from the
# inversion pipeline 'pipelines/with_lens_light/inversion/from_initialize/lens_sie_shear_source_inversion.py'.

# The pipeline is two phases, as follows:

# Phase 1:

# Description: Fits the lens light and mass model as a decomposed profile, using an inversion for the Source.
# Lens Light & Mass: EllipticalSersic
# Lens Mass: SphericalNFW + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Previous Pipelines: with_lens_light/inversion/from_initialize/lens_sie_shear_source_inversion.py
# Prior Passing: Lens Light (variable -> previous pipeline), Lens Mass (default),
#                Source Inversion (variable / constant -> previous pipeline)
# Notes: Uses an interpolation pixel scale for fast power-law deflection angle calculations by default.

# Phase 2:

# Description: Refines the inversion parameters, using a fixed mass model from phase 1.
# Lens Light & Mass: EllipticalSersic
# Lens Mass: SphericalNFW + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Previous Pipelines: None
# Prior Passing: Lens Light & Mass (constant -> phase 1), source inversion (variable -> phase 1)
# Notes: Uses an interpolation pixel scale for fast power-law deflection angle calculations by default.


def make_pipeline(
    pl_align_light_dark_centre=True,
    pl_pixelization=pix.VoronoiMagnification,
    pl_regularization=reg.AdaptiveBrightness,
    phase_folders=None,
    tag_phases=True,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    interp_pixel_scale=0.05,
    inversion_pixel_limit=None,
    cluster_pixel_scale=0.1,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_ldm__lens_sersic_mlr_nfw_shear_source_inversion"
    pipeline_name = tag.pipeline_name_from_name_and_settings(
        pipeline_name=pipeline_name,
        pixelization=pl_pixelization,
        regularization=pl_regularization,
    )

    phase_folders.append(pipeline_name)

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's light and mass and one source galaxy, where we:

    # 1) Pass priors on the lens galaxy's light using the EllipticalSersic of the previous pipeline.
    # 2) Pass priors on the lens galaxy's SphericalNFW mass profile's centre using the EllipticalIsothermal fit of the
    #    previous pipeline, if the NFW centre is a free parameter.
    # 3) Pass priors on the lens galaxy's shear using the ExternalShear fit of the previous pipeline.
    # 4) Pass priors on the source galaxy's inversion using the Pixelization and Regularization of the previous pipeline.

    class LensSourcePhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            ### Lens Light, Sersic -> Sersic ###

            self.lens_galaxies.lens.light = results.from_phase(
                "phase_2_lens_sie_shear_source_inversion"
            ).variable.lens.lens_galaxies.light

            ### Lens Mass, SIE -> Powerlaw ###

            if pl_align_light_dark_centre:

                self.lens_galaxies.lens.dark.centre = self.lens_galaxies.light.centre

            elif not pl_align_light_dark_centre:

                self.lens_galaxies.lens.dark.centre = (
                    results.from_phase("phase_2_lens_sie_shear_source_inversion")
                    .variable_absolute(a=0.05)
                    .lens_galaxies.lens.mass.centre
                )

            ### Lens Shear, Shear -> Shear ###

            self.lens_galaxies.lens.shear = results.from_phase(
                "phase_2_lens_sie_shear_source_inversion"
            ).variable.lens_galaxies.lens.shear

            ### Source Inversion, Inv -> Inv ###

            self.source_galaxies.source = results.from_phase(
                "phase_2_lens_sie_shear_source_inversion"
            ).inversion.constant.source_galaxies.source

    phase1 = LensSourcePhase(
        phase_name="phase_1_lens_sersic_mlr_nfw_shear_source_inversion",
        phase_folders=phase_folders,
        tag_phases=tag_phases,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                light=lp.EllipticalSersic,
                mass=mp.EllipticalPowerLaw,
                shear=mp.ExternalShear,
            )
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(
                redshift=redshift_source,
                pixelization=pl_pixelization,
                regularization=pl_regularization,
            )
        ),
        sub_grid_size=sub_grid_size,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        interp_pixel_scale=interp_pixel_scale,
        inversion_pixel_limit=inversion_pixel_limit,
        cluster_pixel_scale=cluster_pixel_scale,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 75
    phase1.optimizer.sampling_efficiency = 0.2

    phase1 = phase1.extend_with_inversion_phase()

    return pipeline.PipelineImaging(pipeline_name, phase1)



import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging, phase_extensions
from autolens.pipeline import pipeline
from autolens.pipeline import tagging as tag
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp

import os

# In this pipeline, we'll perform an analysis which fits an image with  no lens light, and a source galaxy using a
# parametric light profile, using a power-law mass profile. The pipeline follows on from the initialize pipeline
# ''pipelines/no_lens_light/initialize/lens_sie_shear_source_sersic_from_init.py'.

# The pipeline is one phase, as follows:

# Phase 1:

# Description: Fits the lens mass model as a power-law, using a parametric Sersic light profile for the source.
# Lens Mass: EllipitcalPowerLaw + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: no_lens_light/initialize/lens_sie_shear_source_sersic_from_init.py
# Prior Passing: None
# Notes: Uses an interpolation pixel scale for fast power-law deflection angle calculations.


def make_pipeline(
    phase_folders=None,
    tag_phases=True,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    interp_pixel_scale=0.05,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_pl__lens_pl_shear_source_sersic"

    pipeline_name = tag.pipeline_name_from_name_and_settings(
        pipeline_name=pipeline_name
    )

    phase_folders.append(pipeline_name)

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Set our priors on the lens galaxy mass using the EllipticalIsothermal fit of the previous pipeline, and
    #    source galaxy of the previous pipeline.

    class LensSourcePhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            ### Lens Mass, SIE -> PL ###

            self.lens_galaxies.lens.mass.centre = results.from_phase(
                "phase_1_lens_sie_shear_source_sersic"
            ).variable.lens_galaxies.lens.mass.centre

            self.lens_galaxies.lens.mass.axis_ratio = results.from_phase(
                "phase_1_lens_sie_shear_source_sersic"
            ).variable.lens_galaxies.lens.mass.axis_ratio

            self.lens_galaxies.lens.mass.phi = results.from_phase(
                "phase_1_lens_sie_shear_source_sersic"
            ).variable.lens_galaxies.lens.mass.phi

            self.lens_galaxies.lens.mass.einstein_radius = (
                results.from_phase("phase_1_lens_sie_shear_source_sersic")
                .variable_absolute(a=0.3)
                .lens_galaxies.lens.mass.einstein_radius
            )

            ### Lens Shear, Shear -> Shear ###

            self.lens_galaxies.lens.shear = results.from_phase(
                "phase_1_lens_sie_shear_source_sersic"
            ).variable.lens_galaxies.lens.shear

            ### Source Light, Sersic -> Sersic ###

            self.source_galaxies.source = results.from_phase(
                "phase_1_lens_sie_shear_source_sersic"
            ).variable.source_galaxies.source

    phase1 = LensSourcePhase(
        phase_name="phase_1_lens_pl_shear_source_sersic",
        phase_folders=phase_folders,
        tag_phases=tag_phases,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                mass=mp.EllipticalPowerLaw,
                shear=mp.ExternalShear,
            )
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(redshift=redshift_source, light=lp.EllipticalSersic)
        ),
        sub_grid_size=sub_grid_size,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        interp_pixel_scale=interp_pixel_scale,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 75
    phase1.optimizer.sampling_efficiency = 0.2

    return pipeline.PipelineImaging(pipeline_name, phase1)



import numpy as np
import pytest

from autofit.tools import fit_util
from autolens.data import ccd
from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy as g
from autolens.lens.util import lens_fit_util
from workspace_jam.scraps.stack import (
    ray_tracing_stack,
    lens_data_stack as lds,
    lens_fit_stack,
)
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from test.unit.mock.model.mock_profiles import MockLightProfile


@pytest.fixture(name="lens_data_stack_blur")
def make_li_blur_stack():

    image_0 = np.array(
        [
            [0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 1.0, 0.0],
            [0.0, 1.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0],
        ]
    )
    psf_0 = ccd.PSF(
        array=(np.array([[1.0, 1.0, 1.0], [1.0, 1.0, 1.0], [1.0, 1.0, 1.0]])),
        pixel_scale=1.0,
        renormalize=False,
    )
    ccd_0 = ccd.CCDData(image_0, pixel_scale=1.0, psf=psf_0, noise_map=np.ones((4, 4)))

    mask_0 = np.array(
        [
            [True, True, True, True],
            [True, False, False, True],
            [True, False, False, True],
            [True, True, True, True],
        ]
    )
    mask_0 = msk.Mask(array=mask_0, pixel_scale=1.0)

    image_1 = np.array(
        [
            [0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 1.0, 0.0],
            [0.0, 1.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0],
        ]
    )
    psf_1 = ccd.PSF(
        array=(np.array([[1.0, 1.0, 1.0], [1.0, 1.0, 1.0], [1.0, 1.0, 1.0]])),
        pixel_scale=1.0,
        renormalize=False,
    )
    ccd_1 = ccd.CCDData(image_1, pixel_scale=1.0, psf=psf_1, noise_map=np.ones((4, 4)))

    mask_1 = np.array(
        [
            [True, True, True, True],
            [True, False, False, True],
            [True, False, False, True],
            [True, True, True, True],
        ]
    )
    mask_1 = msk.Mask(array=mask_1, pixel_scale=1.0)

    return lds.LensDataStack(
        ccd_datas=[ccd_0, ccd_1], masks=[mask_0, mask_1], sub_grid_size=1
    )


@pytest.fixture(name="lens_data_stack_manual")
def make_li_manual_stack():

    image_0 = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 2.0, 3.0, 0.0],
            [0.0, 4.0, 5.0, 6.0, 0.0],
            [0.0, 7.0, 8.0, 9.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )
    psf_0 = ccd.PSF(
        array=(np.array([[1.0, 5.0, 9.0], [2.0, 5.0, 1.0], [3.0, 4.0, 0.0]])),
        pixel_scale=1.0,
    )
    ccd_0 = ccd.CCDData(image_0, pixel_scale=1.0, psf=psf_0, noise_map=np.ones((5, 5)))
    mask_0 = msk.Mask(
        array=np.array(
            [
                [True, True, True, True, True],
                [True, False, False, False, True],
                [True, False, False, False, True],
                [True, False, False, False, True],
                [True, True, True, True, True],
            ]
        ),
        pixel_scale=1.0,
    )

    image_1 = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 2.0, 3.0, 0.0],
            [0.0, 4.0, 6.0, 6.0, 0.0],
            [0.0, 7.0, 8.0, 9.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )
    psf_1 = ccd.PSF(
        array=(np.array([[1.0, 1.0, 1.0], [2.0, 1.0, 1.0], [3.0, 1.0, 0.0]])),
        pixel_scale=1.0,
    )
    ccd_1 = ccd.CCDData(image_1, pixel_scale=1.0, psf=psf_1, noise_map=np.ones((5, 5)))
    mask_1 = msk.Mask(
        array=np.array(
            [
                [True, True, True, True, True],
                [True, False, False, False, True],
                [True, False, False, False, True],
                [True, False, False, True, True],
                [True, True, True, True, True],
            ]
        ),
        pixel_scale=1.0,
    )

    return lds.LensDataStack(
        ccd_datas=[ccd_0, ccd_1], masks=[mask_0, mask_1], sub_grid_size=1
    )


class TestLensProfileFit:
    class TestLikelihood:
        def test__image__tracing_fits_data_perfectly__no_psf_blurring__lh_is_noise_normalization(
            self
        ):

            psf_0 = ccd.PSF(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scale=1.0,
            )
            ccd_0 = ccd.CCDData(
                np.ones((3, 3)), pixel_scale=1.0, psf=psf_0, noise_map=np.ones((3, 3))
            )
            mask_0 = msk.Mask(
                array=np.array(
                    [[True, True, True], [True, False, True], [True, True, True]]
                ),
                pixel_scale=1.0,
            )

            psf_1 = ccd.PSF(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scale=1.0,
            )
            ccd_1 = ccd.CCDData(
                np.ones((3, 3)), pixel_scale=1.0, psf=psf_1, noise_map=np.ones((3, 3))
            )
            mask_1 = msk.Mask(
                array=np.array(
                    [[True, True, True], [True, False, True], [True, True, True]]
                ),
                pixel_scale=1.0,
            )

            li_stack = lds.LensDataStack(
                ccd_datas=[ccd_0, ccd_1], masks=[mask_0, mask_1], sub_grid_size=1
            )

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0))
            tracer = ray_tracing_stack.TracerImagePlaneStack(
                lens_galaxies=[g0], image_plane_grid_stacks=li_stack.grid_stacks
            )

            fit = lens_fit_stack.LensProfileFitStack(
                lens_data_stack=li_stack, tracer=tracer
            )
            assert fit.likelihood == 2.0 * -0.5 * np.log(2 * np.pi * 1.0)

        def test__1x2_image__tracing_fits_data_with_chi_sq_5_and_chi_sq_4(self):

            psf_0 = ccd.PSF(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scale=1.0,
            )
            ccd_0 = ccd.CCDData(
                5.0 * np.ones((3, 4)),
                pixel_scale=1.0,
                psf=psf_0,
                noise_map=np.ones((3, 4)),
            )
            ccd_0.image[1, 2] = 4.0
            mask_0 = msk.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            psf_1 = ccd.PSF(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scale=1.0,
            )
            ccd_1 = ccd.CCDData(
                5.0 * np.ones((3, 4)),
                pixel_scale=1.0,
                psf=psf_1,
                noise_map=np.ones((3, 4)),
            )
            mask_1 = msk.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            li_stack = lds.LensDataStack(
                ccd_datas=[ccd_0, ccd_1], masks=[mask_0, mask_1], sub_grid_size=1
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(light_profile=MockLightProfile(value=1.0, size=2))
            tracer = ray_tracing_stack.TracerImagePlaneStack(
                lens_galaxies=[g0], image_plane_grid_stacks=li_stack.grid_stacks
            )

            fit = lens_fit_stack.LensProfileFitStack(
                lens_data_stack=li_stack, tracer=tracer
            )

            assert fit.chi_squared == 25.0 + 32.0
            assert fit.reduced_chi_squared == (25.0 + 32.0) / 2.0
            assert fit.likelihood == -0.5 * (
                (25.0 + 2.0 * np.log(2 * np.pi * 1.0))
                + (32.0 + 2.0 * np.log(2 * np.pi * 1.0))
            )

    class TestCompareToManual:
        def test___manual_image_and_psf(self, lens_data_stack_manual):

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            g1 = g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))

            tracer = ray_tracing_stack.TracerImageSourcePlanesStack(
                lens_galaxies=[g0, g1],
                source_galaxies=[g0],
                image_plane_grid_stacks=lens_data_stack_manual.grid_stacks,
            )

            padded_tracer = ray_tracing_stack.TracerImageSourcePlanesStack(
                lens_galaxies=[g0, g1],
                source_galaxies=[g0],
                image_plane_grid_stacks=lens_data_stack_manual.padded_grid_stacks,
            )

            fit = lens_fit_stack.fit_lens_image_stack_with_tracer(
                lens_data_stack=lens_data_stack_manual,
                tracer=tracer,
                padded_tracer=padded_tracer,
            )

            assert lens_data_stack_manual.noise_maps[0] == pytest.approx(
                fit.noise_maps[0], 1e-4
            )
            assert lens_data_stack_manual.noise_maps[1] == pytest.approx(
                fit.noise_maps[1], 1e-4
            )

            model_image_1d_0 = lens_fit_util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
                unblurred_image_1d=tracer.image_plane_images_1d[0],
                blurring_image_1d=tracer.image_plane_blurring_images_1d[0],
                convolver=lens_data_stack_manual.convolvers_image[0],
            )

            model_image_1d_1 = lens_fit_util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
                unblurred_image_1d=tracer.image_plane_images_1d[1],
                blurring_image_1d=tracer.image_plane_blurring_images_1d[1],
                convolver=lens_data_stack_manual.convolvers_image[1],
            )

            model_image_0 = lens_data_stack_manual.map_to_scaled_arrays[0](
                array_1d=model_image_1d_0
            )
            model_image_1 = lens_data_stack_manual.map_to_scaled_arrays[1](
                array_1d=model_image_1d_1
            )

            assert model_image_0 == pytest.approx(fit.model_images[0], 1e-4)
            assert model_image_1 == pytest.approx(fit.model_images[1], 1e-4)

            residual_map_0 = fit_util.residual_map_from_data_mask_and_model_data(
                data=lens_data_stack_manual.images[0],
                mask=lens_data_stack_manual.masks[0],
                model_data=model_image_0,
            )
            residual_map_1 = fit_util.residual_map_from_data_mask_and_model_data(
                data=lens_data_stack_manual.images[1],
                mask=lens_data_stack_manual.masks[1],
                model_data=model_image_1,
            )

            assert residual_map_0 == pytest.approx(fit.residual_maps[0], 1e-4)
            assert residual_map_1 == pytest.approx(fit.residual_maps[1], 1e-4)

            chi_squared_map_0 = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_0,
                mask=lens_data_stack_manual.masks[0],
                noise_map=lens_data_stack_manual.noise_maps[0],
            )
            chi_squared_map_1 = fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1,
                mask=lens_data_stack_manual.masks[1],
                noise_map=lens_data_stack_manual.noise_maps[1],
            )

            assert chi_squared_map_0 == pytest.approx(fit.chi_squared_maps[0], 1e-4)
            assert chi_squared_map_1 == pytest.approx(fit.chi_squared_maps[1], 1e-4)

            chi_squared_0 = fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_0, mask=lens_data_stack_manual.masks[0]
            )
            noise_normalization_0 = fit_util.noise_normalization_from_noise_map_and_mask(
                noise_map=lens_data_stack_manual.noise_maps[0],
                mask=lens_data_stack_manual.masks[0],
            )
            likelihood_0 = fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared_0, noise_normalization=noise_normalization_0
            )

            chi_squared_1 = fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_1, mask=lens_data_stack_manual.masks[1]
            )
            noise_normalization_1 = fit_util.noise_normalization_from_noise_map_and_mask(
                noise_map=lens_data_stack_manual.noise_maps[1],
                mask=lens_data_stack_manual.masks[1],
            )
            likelihood_1 = fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared_1, noise_normalization=noise_normalization_1
            )

            assert fit.chi_squareds == [chi_squared_0, chi_squared_1]
            assert fit.noise_normalizations == [
                noise_normalization_0,
                noise_normalization_1,
            ]
            assert fit.likelihoods == [likelihood_0, likelihood_1]

            assert likelihood_0 + likelihood_1 == pytest.approx(fit.likelihood, 1e-4)
            assert likelihood_0 + likelihood_1 == fit.figure_of_merit
            blurred_image_of_planes_0 = lens_fit_util.blurred_image_of_planes_from_1d_images_and_convolver(
                total_planes=tracer.total_planes,
                image_plane_image_1d_of_planes=tracer.image_plane_images_1d_of_planes[
                    0
                ],
                image_plane_blurring_image_1d_of_planes=tracer.image_plane_blurring_images_1d_of_planes[
                    0
                ],
                convolver=lens_data_stack_manual.convolvers_image[0],
                map_to_scaled_array=lens_data_stack_manual.map_to_scaled_arrays[0],
            )

            blurred_image_of_planes_1 = lens_fit_util.blurred_image_of_planes_from_1d_images_and_convolver(
                total_planes=tracer.total_planes,
                image_plane_image_1d_of_planes=tracer.image_plane_images_1d_of_planes[
                    1
                ],
                image_plane_blurring_image_1d_of_planes=tracer.image_plane_blurring_images_1d_of_planes[
                    1
                ],
                convolver=lens_data_stack_manual.convolvers_image[1],
                map_to_scaled_array=lens_data_stack_manual.map_to_scaled_arrays[1],
            )

            assert (
                blurred_image_of_planes_0[0] == fit.model_images_of_planes[0][0]
            ).all()
            assert (
                blurred_image_of_planes_0[1] == fit.model_images_of_planes[0][1]
            ).all()
            assert (
                blurred_image_of_planes_1[0] == fit.model_images_of_planes[1][0]
            ).all()
            assert (
                blurred_image_of_planes_1[1] == fit.model_images_of_planes[1][1]
            ).all()

            unmasked_blurred_image_0 = lens_fit_util.unmasked_blurred_image_from_padded_grid_stack_psf_and_unmasked_image(
                padded_grid_stack=lens_data_stack_manual.padded_grid_stacks[0],
                psf=lens_data_stack_manual.psfs[0],
                unmasked_image_1d=padded_tracer.image_plane_images_1d[0],
            )

            unmasked_blurred_image_1 = lens_fit_util.unmasked_blurred_image_from_padded_grid_stack_psf_and_unmasked_image(
                padded_grid_stack=lens_data_stack_manual.padded_grid_stacks[1],
                psf=lens_data_stack_manual.psfs[1],
                unmasked_image_1d=padded_tracer.image_plane_images_1d[1],
            )

            assert (unmasked_blurred_image_0 == fit.unmasked_model_images[0]).all()
            assert (unmasked_blurred_image_1 == fit.unmasked_model_images[1]).all()

            unmasked_blurred_image_of_galaxies_i0 = lens_fit_util.unmasked_blurred_image_of_galaxies_from_psf_and_unmasked_1d_galaxy_images(
                galaxies=padded_tracer.image_plane.galaxies,
                image_plane_image_1d_of_galaxies=padded_tracer.image_plane.image_plane_images_1d_of_galaxies[
                    0
                ],
                padded_grid_stack=lens_data_stack_manual.padded_grid_stacks[0],
                psf=lens_data_stack_manual.psfs[0],
            )

            unmasked_blurred_image_of_galaxies_s0 = lens_fit_util.unmasked_blurred_image_of_galaxies_from_psf_and_unmasked_1d_galaxy_images(
                galaxies=padded_tracer.source_plane.galaxies,
                image_plane_image_1d_of_galaxies=padded_tracer.source_plane.image_plane_images_1d_of_galaxies[
                    0
                ],
                padded_grid_stack=lens_data_stack_manual.padded_grid_stacks[0],
                psf=lens_data_stack_manual.psfs[0],
            )

            unmasked_blurred_image_of_galaxies_i1 = lens_fit_util.unmasked_blurred_image_of_galaxies_from_psf_and_unmasked_1d_galaxy_images(
                galaxies=padded_tracer.image_plane.galaxies,
                image_plane_image_1d_of_galaxies=padded_tracer.image_plane.image_plane_images_1d_of_galaxies[
                    0
                ],
                padded_grid_stack=lens_data_stack_manual.padded_grid_stacks[1],
                psf=lens_data_stack_manual.psfs[1],
            )

            unmasked_blurred_image_of_galaxies_s1 = lens_fit_util.unmasked_blurred_image_of_galaxies_from_psf_and_unmasked_1d_galaxy_images(
                galaxies=padded_tracer.source_plane.galaxies,
                image_plane_image_1d_of_galaxies=padded_tracer.source_plane.image_plane_images_1d_of_galaxies[
                    0
                ],
                padded_grid_stack=lens_data_stack_manual.padded_grid_stacks[1],
                psf=lens_data_stack_manual.psfs[1],
            )

            assert (
                unmasked_blurred_image_of_galaxies_i0[0]
                == fit.unmasked_model_images_of_planes_and_galaxies[0][0][0]
            ).all()
            assert (
                unmasked_blurred_image_of_galaxies_s0[0]
                == fit.unmasked_model_images_of_planes_and_galaxies[0][1][0]
            ).all()
            assert (
                unmasked_blurred_image_of_galaxies_i1[0]
                == fit.unmasked_model_images_of_planes_and_galaxies[1][0][0]
            ).all()
            assert (
                unmasked_blurred_image_of_galaxies_s1[0]
                == fit.unmasked_model_images_of_planes_and_galaxies[1][1][0]
            ).all()



import autofit as af
from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline
from autolens.pipeline import tagging as tag
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg

import os

# In this pipeline, we'll perform a basic analysis which fits two source galaxies using a light profile
# followed by an inversion, where the lens galaxy's light is not present in the image, using two phases:

# Phase 1:

# Description: initialize the lens mass model and source light profile.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: None
# Notes: None

# Phase 2:

# Description: initialize the inversion's pixelization and regularization hyper-parameters, using a previous lens mass
#              model.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Previous Pipelines: initialize/lens_sie_source_sersic_from_init.py
# Prior Passing: Lens Mass (variable -> phase 1).
# Notes: None

# Phase 3:

# Description: Refine the lens mass model and source inversion.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Previous Pipelines: initialize/lens_sie_source_sersic_from_init.py
# Prior Passing: Lens Mass (variable -> phase 1), source inversion (variable -> phase 2).
# Notes: None

# ***NOTE*** Performing this analysis in a pipeline composed of 3 consectutive phases it not ideal, and it is better to
#            breaking the pipeline down into multiple pipelines. This is what is done in the 'pipelines/no_lens_light'
#            folder, using the pipelines:

#            1) initialize/lens_sie_source_sersic_from_init.py (phases 1->3)
#            2) initialize/lens_sie_source_inversion_from_pipeline.py (phases 4->5)

#            See runners/runner_adding_pipelines.py for more details on adding pipelines.


def make_pipeline(
    pl_pixelization=pix.VoronoiMagnification,
    pl_regularization=reg.AdaptiveBrightness,
    phase_folders=None,
    tag_phases=True,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    interp_pixel_scale=None,
    inversion_pixel_limit=None,
    cluster_pixel_scale=0.1,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pl__sie_source_inversion"

    pipeline_name = tag.pipeline_name_from_name_and_settings(
        pipeline_name=pipeline_name
    )

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/phase_name/settings_tag'

    phase_folders.append(pipeline_name)

    # As there is no lens light component, we can use an annular mask throughout this pipeline which removes the
    # central regions of the image.

    def mask_function_annular(image):
        return msk.Mask.circular_annular(
            shape=image.shape,
            pixel_scale=image.pixel_scale,
            inner_radius_arcsec=0.2,
            outer_radius_arcsec=3.3,
        )

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Set our priors on the lens galaxy (y,x) centre such that we assume the image is centred around the lens galaxy.

    class LensSourceX1Phase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            self.lens_galaxies.lens.mass.centre_0 = af.GaussianPrior(
                mean=0.0, sigma=0.1
            )
            self.lens_galaxies.lens.mass.centre_1 = af.GaussianPrior(
                mean=0.0, sigma=0.1
            )

    phase1 = LensSourceX1Phase(
        phase_name="phase_1_source",
        phase_folders=phase_folders,
        tag_phases=tag_phases,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                mass=mp.EllipticalIsothermal,
                shear=mp.ExternalShear,
            )
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(redshift=redshift_source, light=lp.EllipticalSersic)
        ),
        mask_function=mask_function_annular,
        sub_grid_size=sub_grid_size,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        interp_pixel_scale=interp_pixel_scale,
        optimizer_class=af.MultiNest,
    )

    # You'll see these lines throughout all of the example pipelines. They are used to make MultiNest sample the \
    # non-linear parameter space faster (if you haven't already, checkout 'tutorial_7_multinest_black_magic' in
    # 'howtolens/chapter_2_lens_modeling'.

    # Fitting the lens galaxy and source galaxy from uninitialized priors often risks MultiNest getting stuck in a
    # local maxima, especially for the image in this example which actually has two source galaxies. Therefore, whilst
    # I will continue to use constant efficiency mode to ensure fast run time, I've upped the number of live points
    # and decreased the sampling efficiency from the usual values to ensure the non-linear search is robust.

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2

    ### PHASE 1 ###

    # In phase 1, we initialize the inversion's resolution and regularization coefficient, where we:

    # 1) Fix our mass model to the lens galaxy mass-model from phase 3 of the initialize pipeline.
    # 2) Use a circular mask which includes all of the source-galaxy light.

    class InversionPhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            ## Lens Mass, SIE -> SIE ###

            self.lens_galaxies.lens.mass = results.from_phase(
                "phase_1_source"
            ).constant.lens_galaxies.lens.mass

            ## Lens Mass, Shear -> Shear ###

            self.lens_galaxies.lens.shear = results.from_phase(
                "phase_1_source"
            ).constant.lens_galaxies.lens.shear

    phase2 = InversionPhase(
        phase_name="phase_2_initialize_inversion",
        phase_folders=phase_folders,
        tag_phases=tag_phases,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                mass=mp.EllipticalIsothermal,
                shear=mp.ExternalShear,
            )
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(
                redshift=redshift_source,
                pixelization=pl_pixelization,
                regularization=pl_regularization,
            )
        ),
        sub_grid_size=sub_grid_size,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        interp_pixel_scale=interp_pixel_scale,
        inversion_pixel_limit=inversion_pixel_limit,
        cluster_pixel_scale=cluster_pixel_scale,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 20
    phase2.optimizer.sampling_efficiency = 0.8

    phase2 = phase2.extend_with_inversion_phase()

    ### PHASE 3 ###

    # In phase 3, we fit the lens's mass and source galaxy using an inversion, where we:

    # 1) Initialize the priors on the lens galaxy mass using the results of the previous pipeline.
    # 2) Initialize the priors of all source inversion parameters from phase 1.

    class InversionPhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            ## Lens Mass, SIE -> SIE ###

            self.lens_galaxies.lens.mass = results.from_phase(
                "phase_1_source"
            ).variable.lens_galaxies.lens.mass

            ## Lens Mass, Shear -> Shear ###

            self.lens_galaxies.lens.shear = results.from_phase(
                "phase_1_source"
            ).variable.lens_galaxies.lens.shear

            ### Source Inversion, Inv -> Inv ###

            self.source_galaxies.source = results.from_phase(
                "phase_2_initialize_inversion"
            ).inversion.constant.source_galaxies.source

    phase3 = InversionPhase(
        phase_name="phase_3_inversion",
        phase_folders=phase_folders,
        tag_phases=tag_phases,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                mass=mp.EllipticalIsothermal,
                shear=mp.ExternalShear,
            )
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(
                redshift=redshift_source,
                pixelization=pl_pixelization,
                regularization=pl_regularization,
            )
        ),
        sub_grid_size=sub_grid_size,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        interp_pixel_scale=interp_pixel_scale,
        inversion_pixel_limit=inversion_pixel_limit,
        cluster_pixel_scale=cluster_pixel_scale,
        optimizer_class=af.MultiNest,
    )

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 50
    phase3.optimizer.sampling_efficiency = 0.5

    return pipeline.PipelineImaging(pipeline_name, phase1, phase2, phase3)



import os

import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.model.inversion import pixelizations as pix, regularization as reg
from autolens.pipeline.phase import phase_imaging, phase_extensions
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from test.integration import integration_util
from test.simulation import simulation_util

test_type = "lens_and_source_inversion"
test_name = "lens_mass_x1_source_x1_adaptive_weighted"

test_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
output_path = test_path + "output/"
config_path = test_path + "config"
af.conf.instance = af.conf.Config(config_path=config_path, output_path=output_path)


def pipeline():

    integration_util.reset_paths(test_name=test_name, output_path=output_path)
    ccd_data = simulation_util.load_test_ccd_data(
        data_type="no_lens_light_and_source_smooth", data_resolution="LSST"
    )
    pipeline = make_pipeline(test_name=test_name)
    pipeline.run(data=ccd_data)


def make_pipeline(test_name):

    phase1 = phase_imaging.LensSourcePlanePhase(
        phase_name="phase_1",
        phase_folders=[test_type, test_name],
        lens_galaxies=dict(
            lens=gm.GalaxyModel(redshift=0.5, mass=mp.EllipticalIsothermal)
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(redshift=1.0, light=lp.EllipticalSersic)
        ),
        optimizer_class=af.MultiNest,
    )

    class InversionPhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            ## Lens Mass, SIE -> SIE, Shear -> Shear ###

            self.lens_galaxies.lens = results.from_phase(
                "phase_1"
            ).constant.lens_galaxies.lens

    phase2 = InversionPhase(
        phase_name="phase_2",
        phase_folders=[test_type, test_name],
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=0.5, mass=mp.EllipticalIsothermal, shear=mp.ExternalShear
            )
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(
                redshift=1.0,
                pixelization=pix.VoronoiBrightnessImage,
                regularization=reg.AdaptiveBrightness,
            )
        ),
        inversion_pixel_limit=50,
        optimizer_class=af.MultiNest,
    )

    phase2 = phase2.extend_with_multiple_hyper_phases(inversion=True)

    class InversionPhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            ## Lens Mass, SIE -> SIE, Shear -> Shear ###

            self.lens_galaxies.lens = results.from_phase(
                "phase_1"
            ).variable.lens_galaxies.lens

            self.source_galaxies.source = results.from_phase(
                "phase_2"
            ).inversion.constant.source_galaxies.source

    phase3 = InversionPhase(
        phase_name="phase_3",
        phase_folders=[test_type, test_name],
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=0.5, mass=mp.EllipticalIsothermal, shear=mp.ExternalShear
            )
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(
                redshift=1.0,
                pixelization=pix.VoronoiBrightnessImage,
                regularization=reg.AdaptiveBrightness,
            )
        ),
        inversion_pixel_limit=50,
        optimizer_class=af.MultiNest,
    )

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 40
    phase3.optimizer.sampling_efficiency = 0.8

    phase3 = phase3.extend_with_multiple_hyper_phases(inversion=True)

    return pl.PipelineImaging(test_name, phase1, phase2)


if __name__ == "__main__":
    pipeline()



from autofit import conf
import sys

from autolens.data import ccd
from autolens.data.array import mask as msk

import os

# Given your username is where your data is stored, you'll need to put your cosma username here.
cosma_username = "pdtw24"
cosma_path = "/cosma7/data/dp004/dc-nigh1/autolens/"

data_type = "subhalo_jam"


cosma_data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    cosma_path, folder_names=["data", data_type]
)
cosma_output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    cosma_path, folder_names=["output", data_type]
)

workspace_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path, and override the output path with the Cosma path.
conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=cosma_output_path
)

# The fifth line of this batch script - '#SBATCH --array=1-17' is what species this. Its telling Cosma we're going to
# run 17 jobs, and the id's of those jobs will be numbered from 1 to 17. Infact, these ids are passed to this runner,
# and we'll use them to ensure that each jobs loads a different image. Lets get the cosma array id for our job.
cosma_array_id = int(sys.argv[1])

### Subhalo challenge data strings ###

if cosma_array_id == 1:
    pixel_scale = 0.05
    data_name = "large_rein_pl"
    resized_ccd_shape = (400, 400)
elif cosma_array_id == 2:
    pixel_scale = 0.04
    data_name = "large_rein_pl"
    resized_ccd_shape = (400, 400)
elif cosma_array_id == 3:
    pixel_scale = 0.03
    data_name = "large_rein_pl"
    resized_ccd_shape = (500, 500)
elif cosma_array_id == 4:
    pixel_scale = 0.02
    data_name = "large_rein_pl"
    resized_ccd_shape = (700, 700)
elif cosma_array_id == 5:
    pixel_scale = 0.01
    data_name = "large_rein_pl"
    resized_ccd_shape = (1024, 1024)
elif cosma_array_id == 6:
    pixel_scale = 0.05
    data_name = "small_rein_pl"
    resized_ccd_shape = (400, 400)
elif cosma_array_id == 7:
    pixel_scale = 0.04
    data_name = "small_rein_pl"
    resized_ccd_shape = (400, 400)
elif cosma_array_id == 8:
    pixel_scale = 0.03
    data_name = "small_rein_pl"
    resized_ccd_shape = (500, 500)
elif cosma_array_id == 9:
    pixel_scale = 0.02
    data_name = "small_rein_pl"
    resized_ccd_shape = (700, 700)
elif cosma_array_id == 10:
    pixel_scale = 0.01
    data_name = "small_rein_pl"
    resized_ccd_shape = (1024, 1024)

sub_grid_size = 2

data_values = "pixel_scale_" + str(pixel_scale) + "_sub_grid_" + str(sub_grid_size)

data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name, data_values]
)

ccd_data = ccd.load_ccd_data_from_fits(
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    resized_ccd_shape=resized_ccd_shape,
    pixel_scale=pixel_scale,
    resized_psf_shape=(15, 15),
)

mask = msk.load_mask_from_fits(
    mask_path=data_path + "mask_irregular.fits", pixel_scale=pixel_scale
)

positions = ccd.load_positions(positions_path=data_path + "positions.dat")

from workspace_jam.pipelines.advanced.no_lens_light import lens_sie_shear_source_sersic
from workspace_jam.pipelines.advanced.no_lens_light.inversion import (
    lens_sie_shear_source_inversion,
)
from workspace_jam.pipelines.advanced.no_lens_light.power_law.from_inversion import (
    lens_pl_shear_source_inversion,
)
from workspace_jam.pipelines.advanced.no_lens_light.subhalo.from_power_law import (
    lens_pl_shear_subhalo_source_inversion,
)

pipeline_initialize = lens_sie_shear_source_sersic.make_pipeline(
    phase_folders=[data_name, data_values], positions_threshold=0.2
)
pipeline_inversion = lens_sie_shear_source_inversion.make_pipeline(
    phase_folders=[data_name, data_values], positions_threshold=0.2
)
pipeline_power_law = lens_pl_shear_source_inversion.make_pipeline(
    phase_folders=[data_name, data_values], positions_threshold=0.2
)
pipeline_subhalo = lens_pl_shear_subhalo_source_inversion.make_pipeline(
    phase_folders=[data_name, data_values], positions_threshold=0.2
)

pipeline = (
    pipeline_initialize + pipeline_inversion + pipeline_power_law + pipeline_subhalo
)
pipeline.run(data=ccd_data, mask=mask, positions=positions)



import autofit as af
from autolens.pipeline.phase import phase_imaging, phase_extensions
from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy_model as gm
from autolens.data import ccd
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.lens.plotters import lens_fit_plotters

import os

path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))

lens_name = "example_lens"

ccd_data = ccd.load_ccd_data_from_fits(
    image_path=path + "/data/" + lens_name + "/image.fits",
    psf_path=path + "/data/" + lens_name + "/psf.fits",
    noise_map_path=path + "/data/" + lens_name + "/noise_map.fits",
    pixel_scale=0.1,
)

# Create a mask for the data, which we setup as a 3.0" circle.
mask = msk.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=3.0
)

# We model our lens galaxy using a mass profile (a singular isothermal ellipsoid) and our source galaxy
# a light profile (an elliptical Sersic). We load these profiles from the 'light_profiles (lp)' and
# 'mass_profiles (mp)' modules.

# To setup our model galaxies, we use the GalaxyModel class, which represents a galaxy whose parameters
# are variable and fitted for by the analysis. The galaxies are also assigned redshifts.

lens_galaxy_model = gm.GalaxyModel(redshift=0.5, mass=mp.EllipticalIsothermal)
source_galaxy_model = gm.GalaxyModel(redshift=1.0, light=lp.EllipticalSersic)

# To perform the analysis, we set up a phase using the 'phase' module (imported as 'ph').
# A phase takes our galaxy models and fits their parameters using a non-linear search
# (in this case, MultiNest).
phase = phase_imaging.LensSourcePlanePhase(
    lens_galaxies=dict(lens=lens_galaxy_model),
    source_galaxies=dict(source=source_galaxy_model),
    phase_name="example/phase_example",
    optimizer_class=af.MultiNest,
)

# We run the phase on the ccd data, print the results and plot the fit.
result = phase.run(data=ccd_data)
lens_fit_plotters.plot_fit_subplot(fit=result.most_likely_fit)



import copy

import autofit as af
from autolens.model.hyper import hyper_data as hd
from autolens.model.inversion import pixelizations as px
from autolens.model.inversion import regularization as rg
from autolens.pipeline.phase import phase as ph
from .hyper_phase import HyperPhase


# noinspection PyAbstractClass
class VariableFixingHyperPhase(HyperPhase):
    def __init__(
        self,
        phase: ph.Phase,
        hyper_name: str,
        variable_classes=tuple(),
        default_classes=None,
    ):
        super().__init__(phase=phase, hyper_name=hyper_name)
        self.default_classes = default_classes or dict()
        self.variable_classes = variable_classes

    def make_hyper_phase(self):

        phase = super().make_hyper_phase()

        phase.const_efficiency_mode = af.conf.instance.non_linear.get(
            "MultiNest", "extension_inversion_const_efficiency_mode", bool
        )
        phase.optimizer.sampling_efficiency = af.conf.instance.non_linear.get(
            "MultiNest", "extension_inversion_sampling_efficiency", float
        )
        phase.optimizer.n_live_points = af.conf.instance.non_linear.get(
            "MultiNest", "extension_inversion_n_live_points", int
        )

        return phase

    def run_hyper(self, data, results=None, mask=None, positions=None, **kwargs):
        """
        Run the phase, overriding the optimizer's variable instance with one created to
        only fit pixelization hyperparameters.
        """

        variable = results.last.variable.copy_with_fixed_priors(
            results.last.constant, self.variable_classes
        )
        self.add_defaults(variable)

        phase = self.make_hyper_phase()
        phase.optimizer.variable = variable

        return phase.run(data, results=results, mask=mask, positions=positions, **kwargs)

    def add_defaults(self, variable: af.ModelMapper):
        """
        Add default prior models for each of the items in the defaults dictionary.

        Provides a way of specifying new prior models to be included at the top level
        in this phase.

        Parameters
        ----------
        variable
            The variable object to be used in this phase to which default prior
            models are attached.
        """

        # tODO : hasattr did not work for hyper image sky or background, because they were set to None by default.
        # TODO : The getattr below fixes this, but will break for other classes

        for key, value in self.default_classes.items():
            if not hasattr(variable, key) or getattr(variable, key) is None:
                setattr(variable, key, value)


class InversionPhase(VariableFixingHyperPhase):
    """
    Phase that makes everything in the variable from the previous phase equal to the
    corresponding value from the best fit except for variables associated with
    pixelization
    """

    def __init__(
        self,
        phase: ph.Phase,
        variable_classes=(px.Pixelization, rg.Regularization),
        default_classes=None,
    ):
        super().__init__(
            phase=phase,
            variable_classes=variable_classes,
            hyper_name="inversion",
            default_classes=default_classes,
        )

    @property
    def uses_inversion(self):
        return True

    @property
    def uses_hyper_images(self):
        return True


class InversionBackgroundSkyPhase(InversionPhase):
    """
    Phase that makes everything in the variable from the previous phase equal to the
    corresponding value from the best fit except for variables associated with
    pixelization
    """

    def __init__(self, phase: ph.Phase):
        super().__init__(
            phase=phase,
            variable_classes=(px.Pixelization, rg.Regularization, hd.HyperImageSky),
            default_classes={"hyper_image_sky": hd.HyperImageSky},
        )


class InversionBackgroundNoisePhase(InversionPhase):
    """
    Phase that makes everything in the variable from the previous phase equal to the
    corresponding value from the best fit except for variables associated with
    pixelization
    """

    def __init__(self, phase: ph.Phase):
        super().__init__(
            phase=phase,
            variable_classes=(
                px.Pixelization,
                rg.Regularization,
                hd.HyperNoiseBackground,
            ),
            default_classes={"hyper_noise_background": hd.HyperNoiseBackground},
        )


class InversionBackgroundBothPhase(InversionPhase):
    """
    Phase that makes everything in the variable from the previous phase equal to the
    corresponding value from the best fit except for variables associated with
    pixelization
    """

    def __init__(self, phase: ph.Phase):
        super().__init__(
            phase=phase,
            variable_classes=(
                px.Pixelization,
                rg.Regularization,
                hd.HyperImageSky,
                hd.HyperNoiseBackground,
            ),
            default_classes={
                "hyper_image_sky": hd.HyperImageSky,
                "hyper_noise_background": hd.HyperNoiseBackground,
            },
        )
